# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import timedelta

from thrift.protocol.TProtocol import TProtocolException
from thrift.transport.TTransport import TMemoryBuffer
from thrift.transport.TTransport import TTransportException
from tornado import gen
from tornado import locks

from frugal.context import _OPID_HEADER
from frugal.exceptions import TTransportExceptionType
from frugal.tornado.transport.transport import FTransportBase
from frugal.util.headers import _Headers


class FAsyncTransport(FTransportBase):
    """
     FAsyncTransport is an extension of FTransportBase that asynchronous
     frameworks can implement. Implementations need only implement flush to
     send request data and call handle_response when asynchronous responses
     are received.
    """
    def __init__(self, *args, **kwargs):
        super(FAsyncTransport, self).__init__(*args, **kwargs)
        self._futures = {}
        self._futures_lock = locks.Lock()

    @gen.coroutine
    def oneway(self, context, payload):
        self._preflight_request_check(payload)
        try:
            yield gen.with_timeout(
                timedelta(milliseconds=context.timeout),
                self.flush(payload)
            )
        except gen.TimeoutError:
            raise TTransportException(
                type=TTransportExceptionType.TIMED_OUT,
                message='oneway timed out'
            )

    @gen.coroutine
    def request(self, context, payload):
        self._preflight_request_check(payload)
        op_id = str(context._get_op_id())
        future = gen.Future()
        with (yield self._futures_lock.acquire()):
            if op_id in self._futures:
                raise TTransportException(
                    type=TTransportExceptionType.UNKNOWN,
                    message="request already in flight for context"
                )
            self._futures[op_id] = future

        try:
            @gen.coroutine
            def flush_and_wait():
                yield self.flush(payload)
                data = yield future
                raise gen.Return(data)

            data = yield gen.with_timeout(
                timedelta(milliseconds=context.timeout),
                flush_and_wait()
            )
            raise gen.Return(TMemoryBuffer(data))
        except gen.TimeoutError:
            raise TTransportException(
                type=TTransportExceptionType.TIMED_OUT,
                message='request timed out'
            )
        finally:
            with (yield self._futures_lock.acquire()):
                del self._futures[op_id]

    @gen.coroutine
    def flush(self, payload):
        """Flush the payload to the server."""
        raise NotImplementedError('You must override this')

    @gen.coroutine
    def handle_response(self, frame):
        """
        Complete the future associated with the data frame.

        Args:
            frame: The response frame
        """
        if not frame:
            return
        headers = _Headers.decode_from_frame(frame)
        op_id = headers.get(_OPID_HEADER, None)

        if not op_id:
            raise TProtocolException(message="Frame missing op_id")

        with (yield self._futures_lock.acquire()):
            future = self._futures.get(op_id, None)
            if not future:
                return

            future.set_result(frame)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from nats.aio.client import Client
from thrift.transport.TTransport import TTransportException
from thrift.transport.TTransport import TMemoryBuffer

from frugal import _NATS_MAX_MESSAGE_SIZE
from frugal.exceptions import TTransportExceptionType
from frugal.transport import FPublisherTransport
from frugal.transport import FSubscriberTransport
from frugal.transport import FPublisherTransportFactory
from frugal.transport import FSubscriberTransportFactory


class FNatsPublisherTransportFactory(FPublisherTransportFactory):
    """
    FNatsPublisherTransportFactory is used to create
    FNatsPublisherTransports.
    """

    def __init__(self, nats_client: Client):
        self._nats_client = nats_client

    def get_transport(self) -> FPublisherTransport:
        """
        Get a new FNatsPublisherTransport.
        """
        return FNatsPublisherTransport(self._nats_client)


class FNatsPublisherTransport(FPublisherTransport):
    """
    FNatsPublisherTransport is used exclusively for pub/sub scopes.

    Publishers use it to publish to a topic. NATS is used as the
    underlying bus.
    """

    def __init__(self, nats_client: Client):
        super().__init__(_NATS_MAX_MESSAGE_SIZE)
        self._nats_client = nats_client

    async def open(self):
        """
        Open the NATS publisher transport by connected to NATS.
        """
        if not self._nats_client.is_connected:
            raise TTransportException(TTransportExceptionType.NOT_OPEN,
                                      'NATS is not connected')

    async def close(self):
        """
        Close the NATS publisher transport and disconnect from NATS.
        """
        if not self.is_open():
            return

        await self._nats_client.flush()

    def is_open(self) -> bool:
        """
        Check to see if the transport is open.
        """
        return self._nats_client.is_connected

    async def publish(self, topic: str, data):
        """
        Publish a message to NATS on a given topic.

        Args:
            topic: string
            data: bytearray
        """
        if not self.is_open():
            raise TTransportException(TTransportExceptionType.NOT_OPEN,
                                      'Transport is not connected')
        if self._check_publish_size(data):
            raise TTransportException(
                type=TTransportExceptionType.REQUEST_TOO_LARGE,
                message='Message exceeds max message size'
            )
        await self._nats_client.publish('frugal.{0}'.format(topic), data)


class FNatsSubscriberTransportFactory(FSubscriberTransportFactory):
    """
    FNatsSubscriberTransportFactory is used to create
    FNatsSubscriberTransports.
    """

    def __init__(self, nats_client: Client, queue=''):
        self._nats_client = nats_client
        self._queue = queue

    def get_transport(self) -> FSubscriberTransport:
        """
        Get a new FNatsSubscriberTransport.
        """
        return FNatsSubscriberTransport(self._nats_client, self._queue)


class FNatsSubscriberTransport(FSubscriberTransport):
    """
    FSubscriberTransport is used exclusively for pub/sub scopes.
    Subscribers use it to subscribe to a pub/sub topic. Nats is
    used as the underlying bus.
    """

    def __init__(self, nats_client: Client, queue=''):
        self._nats_client = nats_client
        self._queue = queue
        self._is_subscribed = False
        self._sub_id = None

    async def subscribe(self, topic: str, callback):
        """
        Subscribe to the given topic and register a callback to
        invoke when a message is received.

        Args:
            topic: str
            callback: func
        """
        if not self._nats_client.is_connected:
            raise TTransportException(TTransportExceptionType.NOT_OPEN,
                                      'Nats is not connected')
        if self.is_subscribed():
            raise TTransportException(TTransportExceptionType.ALREADY_OPEN,
                                      'Already subscribed to nats topic')

        async def nats_callback(message):
            ret = callback(TMemoryBuffer(message.data[4:]))
            if inspect.iscoroutine(ret):
                ret = await ret
            return ret

        self._sub_id = await self._nats_client.subscribe_async(
            'frugal.{0}'.format(topic),
            queue=self._queue,
            cb=nats_callback,
        )
        self._is_subscribed = True

    async def unsubscribe(self):
        """
        Unsubscribe from the currently subscribed topic.
        """
        await self._nats_client.unsubscribe(self._sub_id)
        self._sub_id = None
        self._is_subscribed = False

    def is_subscribed(self) -> bool:
        """
        Check whether the client is subscribed or not.

        Returns:
            bool
        """
        return self._is_subscribed



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .memory_output_buffer import TMemoryOutputBuffer
from .scope_transport import FPublisherTransport
from .scope_transport import FSubscriberTransport
from .transport import TSynchronousTransport, FTransport
from .transport_factory import FTransportFactory
from .transport_factory import FPublisherTransportFactory
from .transport_factory import FSubscriberTransportFactory

__all__ = [
    'FTransport',
    'TSynchronousTransport',
    'FTransportFactory',
    'TMemoryOutputBuffer',
    'FPublisherTransport',
    'FSubscriberTransport',
    'FPublisherTransportFactory',
    'FSubscriberTransportFactory',
]



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nats.io.utils import new_inbox
from thrift.transport.TTransport import TTransportException
from tornado import gen

from frugal import _NATS_MAX_MESSAGE_SIZE
from frugal.exceptions import TTransportExceptionType
from frugal.tornado.transport import FAsyncTransport

_NOT_OPEN = 'NATS not connected.'
_ALREADY_OPEN = 'NATS transport already open.'


class FNatsTransport(FAsyncTransport):
    """
    FNatsTransport is an extension of FAsyncTransport. This is a "stateless"
    transport in the sense that there is no connection with a server. A request
    is simply published to a subject and responses are received on another
    subject. This assumes requests/responses fit within a single NATS message.
    """

    def __init__(self, nats_client, subject, inbox=""):
        """
        Create a new instance of FNatsTransport

        Args:
            nats_client: connected instance of nats.io.Client
            subject: subject to publish to
        """
        super(FNatsTransport, self).__init__(
            request_size_limit=_NATS_MAX_MESSAGE_SIZE)
        self._nats_client = nats_client
        self._subject = subject
        self._inbox = inbox or new_inbox()
        self._is_open = False
        self._sub_id = None

    def is_open(self):
        return self._is_open and self._nats_client.is_connected

    @gen.coroutine
    def open(self):
        """
        Subscribes to the configured inbox subject.
        """
        if not self._nats_client.is_connected:
            raise TTransportException(
                type=TTransportExceptionType.NOT_OPEN,
                message=_NOT_OPEN)

        elif self.is_open():
            already_open = TTransportExceptionType.ALREADY_OPEN
            raise TTransportException(already_open, _ALREADY_OPEN)

        cb = self._on_message_callback
        inbox = self._inbox
        self._sub_id = yield self._nats_client.subscribe_async(inbox, cb=cb)

        self._is_open = True

    @gen.coroutine
    def _on_message_callback(self, msg):
        yield self.handle_response(msg.data[4:])

    @gen.coroutine
    def close(self):
        """
        Unsubscribes from the inbox subject
        """
        if not self._sub_id:
            return
        yield self._nats_client.flush()
        yield self._nats_client.unsubscribe(self._sub_id)
        self._is_open = False

    @gen.coroutine
    def flush(self, payload):
        """
        Sends the buffered bytes over NATS
        """
        subject = self._subject
        inbox = self._inbox
        yield self._nats_client.publish_request(subject, inbox, payload)



# -*- coding: utf-8 -*-
# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from thrift.protocol import TBinaryProtocol
from thrift.transport import TTransport

from frugal.protocol import FProtocolFactory


def serialize(
        frugal_object,
        protocol_factory=TBinaryProtocol.TBinaryProtocolFactory()):
    """Serialize a frugal entity to bytes."""
    transport = TTransport.TMemoryBuffer()
    fprotocolFactory = FProtocolFactory(protocol_factory)
    protocol = fprotocolFactory.get_protocol(transport)
    frugal_object.write(protocol)
    return transport.getvalue()


def deserialize(
        base,
        buf,
        protocol_factory=TBinaryProtocol.TBinaryProtocolFactory()):
    """Deserialize a frugal object into a base instance of a frugal object."""
    transport = TTransport.TMemoryBuffer(buf)
    fprotocolFactory = FProtocolFactory(protocol_factory)
    protocol = fprotocolFactory.get_protocol(transport)
    base.read(protocol)
    return base



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .async_transport import FAsyncTransport
from .http_transport import FHttpTransport
from .nats_scope_transport import (
    FNatsPublisherTransportFactory,
    FNatsPublisherTransport,
    FNatsSubscriberTransportFactory,
    FNatsSubscriberTransport,
)
from .nats_transport import FNatsTransport
from .transport import FTransportBase


__all__ = [
    'FAsyncTransport',
    'FHttpTransport',
    'FNatsTransport',
    'FNatsPublisherTransport',
    'FNatsPublisherTransportFactory',
    'FNatsSubscriberTransport',
    'FNatsSubscriberTransportFactory',
    'FTransportBase',
]



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class FTransportFactory(object):
    """
    FTransportFactory is responsible for creating new FTransports.
    """

    def get_transport(self, thrift_transport):
        """
        Retuns a new FTransport wrapping the given TTransport.

        Args:
            thrift_transport: TTransport to wrap.
        Returns:
            new FTranpsort
        """
        pass


class FPublisherTransportFactory(object):
    """
    FPublisherTransportFactory is responsible for creating new
    FPublisherTransports.
    """

    def get_transport(self):
        """
        Returns a new FPublisherTransport.
        """
        pass


class FSubscriberTransportFactory(object):
    """
    FSubscriberTransportFactory is responsible for creating new
    FSubscriberTransports.
    """

    def get_transport(self):
        """
        Returns a new FSubscriberTransport.
        """
        pass



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import logging

from thrift.Thrift import TApplicationException
from thrift.transport.TTransport import TMemoryBuffer
from tornado import gen
from tornado.web import RequestHandler

from frugal.transport import TMemoryOutputBuffer

logger = logging.getLogger(__name__)


class FHttpHandler(RequestHandler):
    """
    This class implements a Tornado web server request handler to interface
    with a frugal HTTP client.
    """
    def initialize(self, processor, protocol_factory):
        """

        Args:
            processor: The processor to use to handle requests
            protocol_factory: A protocol factory to serialize/deserialize
                              frugal requests
        """
        self._processor = processor
        self._protocol_factory = protocol_factory

    @gen.coroutine
    def post(self):
        self.set_header('content-type', 'application/x-frugal')

        # check for response size limit
        response_limit = 0
        limit_header_name = "x-frugal-payload-limit"
        if self.request.headers.get(limit_header_name) is not None:
            response_limit = int(self.request.headers[limit_header_name])

        # decode payload and process
        payload = base64.b64decode(self.request.body)
        iprot = self._protocol_factory.get_protocol(
            TMemoryBuffer(payload[4:])
        )

        # TODO could be better with this limit
        otrans = TMemoryOutputBuffer(0)
        oprot = self._protocol_factory.get_protocol(otrans)

        try:
            yield gen.maybe_future(self._processor.process(iprot, oprot))
        except TApplicationException:
            # Continue so the exception is sent to the client
            pass
        except Exception:
            self.send_error(status_code=400)
            return

        # write back response
        output_data = otrans.getvalue()
        if len(output_data) > response_limit > 0:
            self.send_error(status_code=413)
            return

        output_payload = base64.b64encode(output_data)

        self.set_header('content-transfer-encoding', 'base64')
        self.write(output_payload)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webapp2

from frugal.server.http_handler import _FHttpRequest
from frugal.server.http_handler import _FSynchronousHttpRequestHandler


def new_webapp2_handler(processor, protocol_factory):
    """
    Produces a class extending webapp2.RequestHandler that can be used to
    handle frugal HTTP rpc requests.

    Args:
        processor: The processor to use to handle requests.
        protocol_factory: A protocol factory to serialize/deserialize frugal
                          requests.
    """
    handler = _FSynchronousHttpRequestHandler(processor, protocol_factory)

    class FWebapp2Handler(webapp2.RequestHandler):
        """
        FWebapp2Handler uses the webapp2 framework to handle frugal HTTP rpc
        requests.
        """
        def post(self):
            request = _FHttpRequest(
                headers=self.request.headers,
                body=self.request.body,
            )
            response = handler.handle_http_request(request)
            self.response.set_status(response.status_code)
            self.response.headers.update(response.headers)
            self.response.write(response.body)

    return FWebapp2Handler



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class FPublisherTransport(object):
    """
    FPublisherTransport is used exclusively for pub/sub scopes. Publishers use
    it to publish to a topic.
    """

    def __init__(self, max_message_size):
        self._max_message_size = max_message_size

    def open(self):
        """
        Opens the transport for use.
        """
        raise NotImplementedError('You must override this')

    def close(self):
        """
        Closes the transport.
        """
        raise NotImplementedError('You must override this')

    def is_open(self):
        """
        Returns True if the transport is open, False otherwise.
        """
        raise NotImplementedError('You must override this')

    def get_publish_size_limit(self):
        """
        Returns the maximum allowable size of a payload to be published. A
        non-positive number is returned to indicate an unbounded allowable
        size.
        """
        return self._max_message_size

    def publish(self, topic, data):
        """
        Publish sends the given data with the transport to the given topic.
        Implementations of publish should be threadsafe
        """
        raise NotImplementedError('You must override this')

    def _check_publish_size(self, data):
        """
        Returns True if the length of data is over the max message size.
        """
        return len(data) > self._max_message_size > 0


class FSubscriberTransport(object):
    """
    FSubscriberTransport is used exclusively for pub/sub scopes. Subscribers
    use it to subscribe to a pub/sub topic.
    """

    def subscribe(self, topic, callback):
        """
        Subscribes to a pub/sub topic and executes the callback with each
        received message.
        """
        raise NotImplementedError('You must override this')

    def unsubscribe(self):
        """
        Unsubscribes from the current topic.
        """
        raise NotImplementedError('You must override this')

    def remove(self):
        """
        Unsubscribe and removes durably stored information on the broker,
        if applicable.
        """
        return self.unsubscribe()

    def is_subscribed(self):
        """
        Returns True if the transport is subscribed to a topic, False
        otherwise.
        """
        raise NotImplementedError('You must override this')



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import base64

import async_timeout
from aiohttp.client import ClientSession
from thrift.transport.TTransport import TTransportBase
from thrift.transport.TTransport import TMemoryBuffer
from thrift.transport.TTransport import TTransportException

from frugal.aio.transport import FTransportBase
from frugal.context import FContext
from frugal.exceptions import TTransportExceptionType


class FHttpTransport(FTransportBase):
    """
    FHttpTransport is an FTransport that uses http as the underlying transport.
    This allows messages of arbitrary sizes to be sent and received.
    """
    def __init__(self, url, request_capacity=0, response_capacity=0,
                 get_request_headers=None):
        """
        Create an HTTP transport.

        Args:
            url: The url to send requests to.
            request_capacity: The maximum size allowed to be written in a
                              request. Set to 0 for no size restrictions.
            response_capacity: The maximum size allowed to be read in a
                               response. Set to 0 for no size restrictions
            get_request_headers: An optional function that accepts an FContext.
                                 Should return a dictionary of additional
                                 request headers to be appended to the request
        """
        super().__init__(request_capacity)
        self._url = url
        self._get_request_headers = get_request_headers

        self._headers = {
            'content-type': 'application/x-frugal',
            'content-transfer-encoding': 'base64',
            'accept': 'application/x-frugal',
        }
        if response_capacity > 0:
            self._headers['x-frugal-payload-limit'] = str(response_capacity)

    def is_open(self):
        """Always returns True"""
        return True

    async def open(self):
        """No-op"""
        pass

    async def close(self):
        """No-op"""
        pass

    async def oneway(self, context: FContext, payload):
        """
        Write the current buffer. This transport detects oneway requests via
        via the payload size on the server response. Therefore, just call
        through to request.
        """
        await self.request(context, payload)

    async def request(self, context: FContext, payload) -> TTransportBase:
        """
        Write the current buffer payload over the network and return the
        response.
        """
        self._preflight_request_check(payload)
        encoded = base64.b64encode(payload)

        status, text = await self._make_request(context, encoded)
        if status == 413:
            raise TTransportException(
                type=TTransportExceptionType.RESPONSE_TOO_LARGE,
                message='response was too large for the transport'
            )

        if status >= 300:
            raise TTransportException(
                type=TTransportExceptionType.UNKNOWN,
                message='request errored with code {0} and message {1}'.format(
                    status, str(text)
                )
            )

        decoded = base64.b64decode(text)
        if len(decoded) < 4:
            raise TTransportException(type=TTransportExceptionType.UNKNOWN,
                                      message='invalid frame size')

        if len(decoded) == 4:
            if any(decoded):
                raise TTransportException(type=TTransportExceptionType.UNKNOWN,
                                          message='missing data')
            # One-way method, drop response
            return

        return TMemoryBuffer(decoded[4:])

    async def _make_request(self, context: FContext, payload):
        """
        Helper method to make a request over the network.

        Args:
            payload: The data to be sent over the network.
        Return:
            The status code and body of the response.
        Throws:
            TTransportException if the request timed out.
        """
        # construct headers for request
        request_headers = {}
        if self._get_request_headers is not None:
            request_headers = self._get_request_headers(context)
        # apply the default headers so their values cannot be modified
        request_headers.update(self._headers)

        async with ClientSession() as session:
            try:
                with async_timeout.timeout(context.timeout / 1000):
                    async with session.post(self._url,
                                            data=payload,
                                            headers=request_headers) \
                            as response:
                        return response.status, await response.content.read()
            except asyncio.TimeoutError:
                raise TTransportException(
                    type=TTransportExceptionType.TIMED_OUT,
                    message='request timed out'
                )



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from thrift.transport.TTransport import TTransportBase

from frugal.transport import FTransport


class FTransportBase(FTransport):
    """
    FBaseTransport extends FTransport using the async decorators used by all
    asyncio FTransports.
    """

    def is_open(self) -> bool:
        raise NotImplementedError('You must override this')

    async def open(self):
        raise NotImplementedError('You must override this')

    async def close(self):
        raise NotImplementedError('You must override this')

    async def oneway(self, context, payload):
        raise NotImplementedError('You must override this')

    async def request(self, context, payload) -> TTransportBase:
        raise NotImplementedError('You must override this')



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

from frugal.version import __version__

setup(
    name='frugal',
    version=__version__,
    description='Frugal Python Library',
    maintainer='Messaging Team',
    maintainer_email='messaging@workiva.com',
    url='http://github.com/Workiva/frugal',
    packages=find_packages(exclude=('frugal.tests', 'frugal.tests.*')),
    install_requires=[
        'six>=1.10.0,<2',
        'thrift==0.10.0',
        'requests>=2.12.5,<3',
    ],
    extras_require={
        'tornado': ['nats-client==0.7.2'],
        'asyncio': [
            'aiohttp>=3.0.9,<4',
            'aiostomp==1.4.0',
            'asyncio-nats-client==0.8.0',
            'async-timeout>=2.0.1,<4',
        ],
        'gae': ['webapp2==2.5.2'],
    }
)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import warnings

import six


def deprecated(func):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        """
        Return a wrapped deprecated function.
        """
        func_code = six.get_function_code(func)
        warnings.warn_explicit(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            filename=func_code.co_filename,
            lineno=func_code.co_firstlineno + 1
        )
        return func(*args, **kwargs)

    return new_func



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import logging

from aiostomp import AioStomp
from thrift.transport.TTransport import TMemoryBuffer
from thrift.transport.TTransport import TTransportException

from frugal.exceptions import TTransportExceptionType
from frugal.transport import FPublisherTransport
from frugal.transport import FPublisherTransportFactory
from frugal.transport import FSubscriberTransport
from frugal.transport import FSubscriberTransportFactory


_FRUGAL_PREFIX = 'frugal.'
_logger = logging.getLogger(__name__)


class FStompPublisherTransportFactory(FPublisherTransportFactory):
    """
    FStompPublisherTransportFactory is used to create
    FStompPublisherTransports.
    """

    def __init__(self, stomp_client: AioStomp,
                 topic_prefix: str='',
                 max_message_size: int=32 * 1024 * 1024):
        """
        Create an FStompPublisherTransportFactory.

        :param stomp_client: The stomp client to use.
        :param topic_prefix: A string to be added to the beginning of the
                             constructed topic.
        :param max_message_size: The maximum allowed size for a published
                                 message.
        """
        self._stomp_client = stomp_client
        self._topic_prefix = topic_prefix
        self._max_message_size = max_message_size

    def get_transport(self) -> FPublisherTransport:
        """
        Get a new FStompPublisherTransport.
        """
        return _FStompPublisherTransport(
            self._stomp_client, self._topic_prefix, self._max_message_size)


class _FStompPublisherTransport(FPublisherTransport):
    """
    FStompPublisherTransport is used exclusively for pub/sub scopes.
    Publishers use it to publish to a topic. Messaging brokers that support
    stomp protocol can be used as the underlying bus.
    """

    def __init__(self, stomp_client: AioStomp,
                 topic_prefix: str='',
                 max_message_size: int=32 * 1024 * 1024):
        super().__init__(max_message_size)
        self._stomp_client = stomp_client
        self._topic_prefix = topic_prefix
        self._is_open = False

    async def open(self):
        """
        No-op. Client connection should be established outside of frugal.
        """
        self._is_open = True

    async def close(self):
        """
        No-op. Client close should be handled outside of frugal.
        :return:
        """
        self._is_open = False

    def is_open(self) -> bool:
        """
        No-op.
        """
        return self._is_open

    async def publish(self, topic: str, data):
        """
        Publish a message to stomp broker on a given topic.

        Args:
            topic: string
            data: bytearray
        """
        if self._check_publish_size(data):
            raise TTransportException(
                type=TTransportExceptionType.REQUEST_TOO_LARGE,
                message='Message exceeds max message size'
            )

        destination = '/topic/{topic_prefix}{frugal_prefix}{topic}'.format(
            topic_prefix=self._topic_prefix,
            frugal_prefix=_FRUGAL_PREFIX,
            topic=topic
        )
        _logger.debug(
            'publishing stomp message on topic {}'.format(destination))
        self._stomp_client.send(
            destination,
            data,
            headers={'persistent': 'true',
                     'content-type': 'application/octet-stream'}
        )
        _logger.debug(
            'published stomp message on topic {}'.format(destination))


class FStompSubscriberTransportFactory(FSubscriberTransportFactory):
    """
    FStompSubscriberTransportFactory is used to create
    FStompSubscriberTransports.
    """

    def __init__(self, stomp_client: AioStomp, topic_prefix: str='',
                 use_queue: bool=False):
        """
        Create a new FStompSubscriberTransportFactory.

        :param stomp_client: The stomp client to use.
        :param topic_prefix: A string to be added to the beginning of the
                             constructed topic.
        :param use_queue: Whether the subscriber should use queues or topics.
                          This could be used to take advantage of virtual
                          topics in activemq.
        """
        self._stomp_client = stomp_client
        self._topic_prefix = topic_prefix
        self._use_queue = use_queue

    def get_transport(self) -> FSubscriberTransport:
        """
        Get a new FStompSubscriberTransport.
        """
        return _FStompSubscriberTransport(
            self._stomp_client, self._topic_prefix, self._use_queue)


class _FStompSubscriberTransport(FSubscriberTransport):
    """
    FStompSubscriberTransport is used exclusively for pub/sub scopes.
    Subscribers use it to subscribe to a pub/sub topic. Messaging brokers that
    support stomp protocol can be used as the underlying bus.
    """

    def __init__(self, stomp_client: AioStomp, topic_prefix: str,
                 use_queue: bool=True):
        self._stomp_client = stomp_client
        self._topic_prefix = topic_prefix
        self._use_queue = use_queue
        self._sub = None
        self._destination = None

    async def subscribe(self, topic: str, callback):
        """
        Subscribe to the given topic and register a callback to
        invoke when a message is received.

        If an exception is raised by the provided callback, the message will
        not be acked with the broker. This behaviour allows the message to be
        redelivered and processing to be attempted again. If an exception is
        not raised by the provided callback, the message will be acked. This is
        used if processing succeeded, or if it's apparent processing will never
        succeed, as the message won't continue to be redelivered.

        Args:
            topic: str
            callback: func
        """
        if self.is_subscribed():
            raise TTransportException(
                TTransportExceptionType.ALREADY_OPEN,
                'stomp connection already subscribed to topic')

        if not topic:
            raise TTransportException(
                TTransportExceptionType.UNKNOWN,
                'stomp transport cannot subscribe to empty topic')

        self._destination = \
            '/{type}/{topic_prefix}{frugal_prefix}{topic}'.format(
                type='queue' if self._use_queue else 'topic',
                topic_prefix=self._topic_prefix,
                frugal_prefix=_FRUGAL_PREFIX,
                topic=topic
            )

        async def msg_handler(frame, _):
            _logger.debug('received stomp message on topic \'{}\''.
                          format(self._destination))
            try:
                ret = callback(TMemoryBuffer(frame.body[4:]))
                if inspect.iscoroutine(ret):
                    await ret
                _logger.debug(
                    'finished processing stomp message from topic \'{}\''.
                    format(self._destination))
                # aiostomp acks message automatically in client-individual mode
                # as long as handler function returns non-falsy value
                return True
            except Exception:
                # catch exceptions so the stomp library doesn't barf and keeps
                # processing messages
                _logger.exception(
                    'unable to process stomp message from topic \'{}\''.
                    format(self._destination)
                )

        self._sub = self._stomp_client.subscribe(
            self._destination, ack='client-individual', handler=msg_handler)

    async def unsubscribe(self):
        """
        Unsubscribe from the currently subscribed topic.
        """
        if not self.is_subscribed():
            return

        self._stomp_client.unsubscribe(self._sub)
        self._sub = None

    def is_subscribed(self) -> bool:
        """
        Check whether the client is subscribed or not.
        """
        return bool(self._sub)
